#include "../DivideAndConquer.hpp"

#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include <iterator>
#include <vector>

SCENARIO ("median3 works", "[median3]")
{
  GIVEN ("A vector with at least 3 elements")
  {
    std::vector<int> v{1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
    std::vector<int> copy (v);
    auto first = v.begin ();
    auto last = --v.end ();
    auto mid = v.begin () + v.size () / 2;
    WHEN ("The ordering is [low] <= [mid] <= [high]")
    {
      AND_WHEN ("We call median3")
      {
        auto result = SortUtils::median3 (v.begin (), v.end ());
        THEN ("[1] nothing should be swapped")
        {
          REQUIRE (*first <= *mid);
          REQUIRE (*mid <= *last);
          REQUIRE (v == copy);
        }
        THEN ("[0.55] the correct iterator is returned")
        {
          REQUIRE (result == mid);
        }
      }
    }
    WHEN ("The ordering is [low] <= [high] <= [mid]")
    {
      std::iter_swap (mid, last);
      AND_WHEN ("We call median3")
      {
        auto result = SortUtils::median3 (v.begin (), v.end ());
        THEN ("[1.1] values should be swapped")
        {
          REQUIRE (*first <= *mid);
          REQUIRE (*mid <= *last);
          REQUIRE (v == copy);
        }
        THEN ("[0.55] the correct iterator is returned")
        {
          REQUIRE (result == mid);
        }
      }
    }
    WHEN ("The ordering is [mid] <= [low] <= [high]")
    {
      std::iter_swap (mid, last);
      AND_WHEN ("We call median3")
      {
        auto result = SortUtils::median3 (v.begin (), v.end ());
        THEN ("[1.1] values should be swapped")
        {
          REQUIRE (*first <= *mid);
          REQUIRE (*mid <= *last);
          REQUIRE (v == copy);
        }
        THEN ("[0.55] the correct iterator is returned")
        {
          REQUIRE (result == mid);
        }
      }
    }
    WHEN ("The ordering is [mid] <= [high] <= [low]")
    {
      std::iter_swap (mid, last);
      std::iter_swap (first, last);
      AND_WHEN ("We call median3")
      {
        auto result = SortUtils::median3 (v.begin (), v.end ());
        THEN ("[1.2] values should be swapped")
        {
          REQUIRE (*first <= *mid);
          REQUIRE (*mid <= *last);
          REQUIRE (v == copy);
        }
        THEN ("[0.55] the correct iterator is returned")
        {
          REQUIRE (result == mid);
        }
      }
    }
    WHEN ("The ordering is [high] <= [low] <= [mid]")
    {
      std::iter_swap (first, last);
      std::iter_swap (mid, last);
      AND_WHEN ("We call median3")
      {
        auto result = SortUtils::median3 (v.begin (), v.end ());
        THEN ("[1.2] values should be swapped")
        {
          REQUIRE (*first <= *mid);
          REQUIRE (*mid <= *last);
          REQUIRE (v == copy);
        }
        THEN ("[0.55] the correct iterator is returned")
        {
          REQUIRE (result == mid);
        }
      }
    }
    WHEN ("The ordering is [high] <= [mid] <= [low]")
    {
      std::iter_swap (first, last);
      AND_WHEN ("We call median3")
      {
        auto result = SortUtils::median3 (v.begin (), v.end ());
        THEN ("[1.1] values should be swapped")
        {
          REQUIRE (*first <= *mid);
          REQUIRE (*mid <= *last);
          REQUIRE (v == copy);
        }
        THEN ("[0.55] the correct iterator is returned")
        {
          REQUIRE (result == mid);
        }
      }
    }
  }
}

SCENARIO ("merge works", "[merge]")
{
  GIVEN ("Two sorted vectors")
  {
    std::vector<int> a (10);
    std::vector<int> b (4);
    std::vector<int> out (a.size () + b.size ());
    std::vector<int> copy_out (out);
    auto expected = out.end ();
    WHEN ("All elements in Range1 are less than all elements in Range2")
    {
      std::iota (a.begin (), a.end (), 0);
      std::iota (b.begin (), b.end (), a.size () * 2);
      AND_WHEN ("We call merge")
      {
        auto result = SortUtils::merge (a.begin (), a.end (), b.begin (),
                                        b.end (), out.begin ());
        std::merge (a.begin (), a.end (), b.begin (), b.end (),
                    copy_out.begin ());
        THEN ("[2] We expect all elements to be merged to out")
        {
          REQUIRE (out == copy_out);
        }
        THEN ("[1] The return value is correct")
        {
          REQUIRE (result == expected);
        }
      }
    }
    WHEN ("All elements in Range2 are less than all elements in Range1")
    {
      std::iota (a.begin (), a.end (), b.size () * 2);
      std::iota (b.begin (), b.end (), 0);
      AND_WHEN ("We call merge")
      {
        auto result = SortUtils::merge (a.begin (), a.end (), b.begin (),
                                        b.end (), out.begin ());
        std::merge (a.begin (), a.end (), b.begin (), b.end (),
                    copy_out.begin ());
        THEN ("[2] We expect all elements to be merged to out")
        {
          REQUIRE (out == copy_out);
        }
        THEN ("[1] The return value is correct")
        {
          REQUIRE (result == expected);
        }
      }
    }
    WHEN ("All elements are mixed between Range1 and Range2")
    {
      size_t i = 0;
      for (auto& v : a)
      {
        v = 2 * i++ + 1;
      }
      i = 0;
      for (auto& v : b)
      {
        v = 2 * i++;
      }
      AND_WHEN ("We call merge")
      {
        auto result = SortUtils::merge (a.begin (), a.end (), b.begin (),
                                        b.end (), out.begin ());
        std::merge (a.begin (), a.end (), b.begin (), b.end (),
                    copy_out.begin ());
        THEN ("[3] We expect all elements to be merged to out")
        {
          REQUIRE (out == copy_out);
        }
        THEN ("[1] The return value is correct")
        {
          REQUIRE (result == expected);
        }
      }
    }
  }
}

SCENARIO ("partition works", "[partition]")
{
  int const pivot = GENERATE (range (1, 11));
  std::vector<int> v;
  v.reserve (100);
  const int seed = 1337;
  std::minstd_rand rng (seed);
  for (int i = 1; i <= 10; ++i)
  {
    v.insert (v.end (), 1 + rng () % 5, i);
  }
  std::shuffle (v.begin (), v.end (), rng);
  GIVEN ("A vector with some duplicates")
  {
    std::vector<int> copy (v);
    WHEN ("We pick a random pivot")
    {
      INFO ("Before partition:");
      CAPTURE (v);
      auto result = SortUtils::partition (v.begin (), v.end (), pivot);
      INFO ("After partition:");
      auto dup_begin = result.first;
      auto greater_begin = result.second;
      CAPTURE (pivot, v);
      THEN ("[2] the vector is properly partitioned")
      {
        auto iter = v.begin ();
        while (iter != v.end () && *iter < pivot)
          ++iter;
        {
          INFO (
            "The first of the pair should be where iter is no longer < pivot");
          REQUIRE (std::distance (v.begin (), iter) ==
                   std::distance (v.begin (), dup_begin));
        }
        while (iter != v.end () && *iter == pivot)
          ++iter;
        {
          INFO (
            "The second of the pair should be where iter is no longer == "
            "pivot");
          REQUIRE (std::distance (v.begin (), iter) ==
                   std::distance (v.begin (), greater_begin));
        }
        while (iter != v.end () && *iter > pivot)
          ++iter;
        {
          INFO ("No elements should exist after testing > pivot");
          REQUIRE (iter == v.end ());
        }
        std::sort (v.begin (), v.end ());
        std::sort (copy.begin (), copy.end ());
        INFO ("There were no loss of elements");
        REQUIRE (v == copy);
      }
    }
  }
}

SCENARIO ("nth_element works", "[nth_element]")
{
  std::vector<int> v (40);
  std::iota (v.begin (), v.end (), 1);
  std::shuffle (v.begin (), v.end (), std::minstd_rand{2047});

  GIVEN ("A vector of some size")
  {
    WHEN ("We call nth_element")
    {
      int const index = GENERATE (0, 39, take (13, random (1, 40)));
      INFO ("Before call to nth_element");
      CAPTURE (v, index);
      auto copy = v;
      auto result = SortUtils::nth_element (v.begin (), v.end (), index);
      THEN ("[1] We get the correct value")
      {
        INFO ("After call to nth_element");
        CAPTURE (v);
        {
          INFO ("the value the iterator points to is correct");
          REQUIRE (*result == (index + 1));
        }
        {
          INFO ("the index of the iterator returned is correct");
          REQUIRE (std::distance (v.begin (), result) + 1 == *result);
        }
        std::sort (v.begin (), v.end ());
        std::sort (copy.begin (), copy.end ());
        {
          INFO ("No elements were lost");
          REQUIRE (v == copy);
        }
      }
    }
  }
}

template<typename T>
std::vector<T>
sortChunksOf5 (std::vector<T> data)
{
  for (auto i = data.begin (); i < data.end (); i += 5)
  {
    std::sort (i, std::min ({i + 5, data.end ()}));
  }
  return data;
}

template<typename T>
std::vector<T>
mediansOf5 (std::vector<T> const& data)
{
  std::vector<T> result;
  for (size_t i = 2; i < data.size (); i += 5)
  {
    result.push_back (data[i]);
  }
  return result;
}

SCENARIO ("median5 works", "[median5]")
{
  GIVEN ("A vector with a small size")
  {
    std::vector<int> v (5);
    std::iota (v.begin (), v.end (), 1);
    std::minstd_rand rng (123);
    std::shuffle (v.begin (), v.end (), rng);
    std::vector<int> copy (v);
    WHEN ("We call median5")
    {
      auto result = SortUtils::median5 (v.begin (), v.end ());
      THEN ("[5] we successfully determine the median")
      {
        CAPTURE (*result, v);
        {
          INFO ("data should be sorted with insertion sort for size <= 5");
          REQUIRE (std::is_sorted (v.begin (), v.end ()));
        }
        std::sort (copy.begin (), copy.end ());
        {
          INFO ("no elements should be lost");
          REQUIRE (
            std::equal (v.begin (), v.end (), copy.begin (), copy.end ()));
        }
        {
          INFO ("the iterator points to the correct location");
          REQUIRE (result == v.begin () + v.size () / 2);
        }
        {
          INFO ("the iterator has the correct value");
          REQUIRE (*result == copy[copy.size () / 2]);
        }
      }
    }
  }
  GIVEN ("A vector with a medium size")
  {
    std::vector<int> v (25);
    std::iota (v.begin (), v.end (), 1);
    std::minstd_rand rng (123);
    std::shuffle (v.begin (), v.end (), rng);
    std::vector<int> copy (v);
    WHEN ("We call median5")
    {
      auto result = SortUtils::median5 (v.begin (), v.end ());
      auto sortedChunks = sortChunksOf5 (copy);
      auto mediansOfChunks = mediansOf5 (sortedChunks);

      auto sortedChunks2 = sortChunksOf5 (mediansOfChunks);
      auto mediansOfChunks2 = mediansOf5 (sortedChunks2);

      auto median = mediansOfChunks2[0];

      THEN ("[5] we successfully determine the median")
      {
        CAPTURE (*result, v);

        {
          INFO ("the iterator points to the correct location");
          REQUIRE (result == v.begin () + 2);
        }
        {
          INFO ("the iterator has the correct value");
          REQUIRE (*result == median);
        }
        {
          INFO ("the first 5 elements are the medians of the 5 groups");
          std::vector<int> firstFiveSorted (v.begin (), v.begin () + 5);
          std::sort (firstFiveSorted.begin (), firstFiveSorted.end ());
          auto mediansOfChunksSorted = mediansOfChunks;
          std::sort (mediansOfChunksSorted.begin (),
                     mediansOfChunksSorted.end ());
          REQUIRE (mediansOfChunksSorted == firstFiveSorted);
        }
        {
          INFO ("the first 5 elements should be sorted (base case)");
          REQUIRE (std::is_sorted (v.begin (), v.begin () + 5));
        }
        {
          INFO ("No elements should be lost");
          auto vcopy = v;
          std::sort (vcopy.begin (), vcopy.end ());
          auto orig = copy;
          std::sort (orig.begin (), orig.end ());
          REQUIRE (std::equal (vcopy.begin (), vcopy.end (), orig.begin (),
                               orig.end ()));
        }
      }
    }
  }
  GIVEN ("A vector with a large size")
  {
    std::vector<int> v (125);
    std::iota (v.begin (), v.end (), 1);
    std::minstd_rand rng (123);
    std::shuffle (v.begin (), v.end (), rng);
    std::vector<int> copy (v);
    INFO ("Before operation");
    CAPTURE (v);
    WHEN ("We call median5")
    {

      INFO ("Expected");
      auto result = SortUtils::median5 (v.begin (), v.end ());
      auto mediansOfChunks = mediansOf5 (sortChunksOf5 (copy));
      auto mediansOfChunks2 = mediansOf5 (sortChunksOf5 (mediansOfChunks));
      auto mediansOfChunks3 = mediansOf5 (sortChunksOf5 (mediansOfChunks2));
      CAPTURE (mediansOfChunks, mediansOfChunks2, mediansOfChunks3);

      INFO ("After operation");
      auto median = mediansOfChunks3[0];

      THEN ("[5] we successfully determine the median")
      {
        CAPTURE (*result, v);
        {
          INFO ("the iterator points to the correct location");
          REQUIRE (result == v.begin () + 2);
        }
        {
          INFO ("the iterator has the correct value");
          REQUIRE (*result == median);
        }
        {
          INFO ("the first 5 elements are the medians of the median groups");
          std::vector<int> firstFiveSorted (v.begin (), v.begin () + 5);
          std::sort (firstFiveSorted.begin (), firstFiveSorted.end ());
          auto mediansOfChunksSorted = mediansOfChunks2;
          std::sort (mediansOfChunksSorted.begin (),
                     mediansOfChunksSorted.end ());
          REQUIRE (mediansOfChunksSorted == firstFiveSorted);
        }
        {
          INFO ("the first 5 elements should be sorted (base case)");
          REQUIRE (std::is_sorted (v.begin (), v.begin () + 5));
        }
        {
          INFO ("No elements should be lost");
          auto vcopy = v;
          std::sort (vcopy.begin (), vcopy.end ());
          auto orig = copy;
          std::sort (orig.begin (), orig.end ());
          REQUIRE (std::equal (vcopy.begin (), vcopy.end (), orig.begin (),
                               orig.end ()));
        }
      }
    }
  }
  GIVEN ("A vector with a huge size")
  {
    std::vector<int> v (3125);
    std::iota (v.begin (), v.end (), 1);
    std::minstd_rand rng (123);
    std::shuffle (v.begin (), v.end (), rng);
    std::vector<int> copy (v);
    WHEN ("We call median5")
    {
      INFO ("Expected");
      auto result = SortUtils::median5 (v.begin (), v.end ());
      auto mediansOfChunks = mediansOf5 (sortChunksOf5 (copy));
      auto mediansOfChunks2 = mediansOf5 (sortChunksOf5 (mediansOfChunks));
      auto mediansOfChunks3 = mediansOf5 (sortChunksOf5 (mediansOfChunks2));
      auto mediansOfChunks4 = mediansOf5 (sortChunksOf5 (mediansOfChunks3));
      auto mediansOfChunks5 = mediansOf5 (sortChunksOf5 (mediansOfChunks4));

      auto median = mediansOfChunks5[0];

      THEN ("[5] we successfully determine the median")
      {
        CAPTURE (*result);
        {
          INFO ("the iterator points to the correct location");
          REQUIRE (std::distance (v.begin (), result) == 2);
        }
        {
          INFO ("the iterator has the correct value");
          REQUIRE (*result == median);
        }
        {
          INFO ("the first 5 elements are the medians of the median groups");
          std::vector<int> firstFiveSorted (v.begin (), v.begin () + 5);
          std::sort (firstFiveSorted.begin (), firstFiveSorted.end ());
          auto mediansOfChunksSorted = mediansOfChunks4;
          std::sort (mediansOfChunksSorted.begin (),
                     mediansOfChunksSorted.end ());
          REQUIRE (mediansOfChunksSorted == firstFiveSorted);
        }
        {
          INFO ("the first 5 elements should be sorted (base case)");
          REQUIRE (std::is_sorted (v.begin (), v.begin () + 5));
        }
        {
          INFO ("No elements should be lost");
          auto vcopy = v;
          std::sort (vcopy.begin (), vcopy.end ());
          auto orig = copy;
          std::sort (orig.begin (), orig.end ());
          REQUIRE (std::equal (vcopy.begin (), vcopy.end (), orig.begin (),
                               orig.end ()));
        }
      }
    }
  }
}
